@@ -1,7 +1,7 @@
 import logging
 import re
 
-from collections import namedtuple, defaultdict
+from collections import defaultdict, deque, namedtuple
 from concurrent.futures import as_completed
 from itertools import chain
 
@@ -26,6 +26,10 @@ class RepoMissing(Exception):
     pass
 
 
+class ConfigMissing(Exception):
+    pass
+
+
 RepoSet = namedtuple('RepoSet', ['rpm', 'source', 'debug'])
 
 
@@ -80,38 +84,141 @@ class UbiRepoSet(object):
 
 class UbiPopulate(object):
     def __init__(self, pulp_hostname, pulp_auth, dry_run, ubiconfig_filename_list=None,
-                 ubiconfig_dir_or_url=None, insecure=False, workers_count=4,
-                 output_repos=None):
+                 ubiconfig_dir_or_url=None, insecure=False, workers_count=4, output_repos=None,
+                 **kwargs):
 
-        self.ubiconfig_list = self._load_ubiconfig(ubiconfig_filename_list,
-                                                   ubiconfig_dir_or_url)
         self.pulp = Pulp(pulp_hostname, pulp_auth, insecure)
         self.dry_run = dry_run
         self.output_repos = output_repos
         self._executor = Executors.thread_pool(max_workers=workers_count).with_retry()
+        self.ubiconfig_list = self._load_ubiconfig(ubiconfig_filename_list, ubiconfig_dir_or_url,
+                                                   content_sets=kwargs.get('content_sets', None),
+                                                   repo_ids=kwargs.get('repo_ids', None))
+        self.ubiconfig_map = self._create_config_map()
 
-    def _load_ubiconfig(self, filenames, ubiconfig_dir_or_url):
+    def _load_ubiconfig(self, filenames, ubiconfig_dir_or_url, content_sets=None, repo_ids=None):
         loader = ubiconfig.get_loader(ubiconfig_dir_or_url)
         ubi_conf_list = []
+
         if filenames:
             ubi_conf_list.extend(loader.load(filename) for filename in filenames)
         else:
             ubi_conf_list.extend(loader.load_all())
 
-        return ubi_conf_list
+        return self._filter_ubi_conf_list(ubi_conf_list, content_sets, repo_ids)
+
+    def _filter_ubi_conf_list(self, config_list, content_sets, repo_ids):
+        """
+        Reduces the list of UBI configurations to only those matching
+        provided content sets and/or repo IDs.
+        """
+
+        filtered_conf_list = []
+
+        content_sets = content_sets or []
+        repo_ids = repo_ids or []
+
+        if not content_sets and not repo_ids:
+            return config_list
+
+        fts = [self._executor.submit(self.pulp.search_repo_by_id, r) for r in repo_ids]
+        for repo_list in [ft.result() for ft in fts]:
+            content_sets.extend(repo.content_set for repo in repo_list)
+
+        for conf in config_list:
+            for label in [
+                    conf.content_sets.rpm.input, conf.content_sets.rpm.output,
+                    conf.content_sets.srpm.input, conf.content_sets.srpm.output,
+                    conf.content_sets.debuginfo.input, conf.content_sets.debuginfo.output,
+            ]:
+                if label in content_sets:
+                    filtered_conf_list.append(conf)
+                    break
+
+        return filtered_conf_list
+
+    def _create_config_map(self):
+        """Create a config map from self.ubiconfig_list, it has the form in:
+            {
+                "7.7":
+                    {
+                        "config_filename1": config1,
+                        "config_filename2": config2,
+                        ...,
+                    },
+                "8.1":
+                    {
+                        "config_filename1": config1,
+                        ...,
+                    },
+                ....
+            }
+        """
+
+        config_map = {}
+        for config in self.ubiconfig_list:
+            config_map.setdefault(config.version, {})\
+                .setdefault(config.file_name, config)
+
+        return config_map
 
     def populate_ubi_repos(self):
         out_repos = set()
+        used_content_sets = set()
+        # since repos are searched by content sets, same repo could be searched and populated
+        # multiple times, to avoid that, cache the content sets already used and skip the config
+        # whose content sets are all in the cache
+
+        for config in sorted(self.ubiconfig_list, key=str):
+            content_sets = [
+                config.content_sets.rpm.output,
+                config.content_sets.srpm.output,
+                config.content_sets.debuginfo.output,
+            ]
+
+            to_use = [cs for cs in content_sets if cs not in used_content_sets]
+            if to_use:
+                for cs in to_use:
+                    used_content_sets.add(cs)
+            else:
+                _LOG.debug("Skipping %s, since it's been used already", config.file_name)
+                continue
 
-        for config in self.ubiconfig_list:
             try:
-                repo_pairs = self._get_input_and_output_repo_pairs(config)
+                repo_pairs = self._get_ubi_repo_sets(config)
+
             except RepoMissing:
                 _LOG.warning("Skipping current content triplet, some repos are missing")
                 continue
 
             for repo_set in repo_pairs:
-                UbiPopulateRunner(self.pulp, repo_set, config, self.dry_run,
+                ubi_config_version = repo_set.out_repos.rpm.ubi_config_version
+                platform_full_version = repo_set.out_repos.rpm.platform_full_version
+                platform_major_version = repo_set.out_repos.rpm.platform_major_version
+                # get the right config file by ubi_config_version attr, if it's None,
+                # then it's not a mainline repo, use platform_full_version instead.
+                # config file could also be missing for specific version, then the
+                # default config file will be used.
+                version = ubi_config_version or platform_full_version
+                right_config = self.ubiconfig_map\
+                    .get(str(version), {})\
+                    .get(config.file_name)\
+                    or self.ubiconfig_map\
+                        .get(str(platform_major_version), {})\
+                        .get(config.file_name)
+
+                # if config file is missing from wanted version, as well as default
+                # branch, raise exception
+                if not right_config:
+                    _LOG.error(
+                        'Config file %s missing from %s and default %s branches',
+                        config.file_name,
+                        version,
+                        platform_full_version,
+                    )
+                    raise ConfigMissing()
+
+                UbiPopulateRunner(self.pulp, repo_set, right_config, self.dry_run,
                                   self._executor).run_ubi_population()
 
                 out_repos.update(repo_set.get_output_repo_ids())
@@ -121,50 +228,72 @@ class UbiPopulate(object):
                 for repo in out_repos:
                     f.write(repo.strip() + '\n')
 
-    def _get_input_and_output_repo_pairs(self, ubiconfig_item):
+    def _get_ubi_repo_sets(self, ubi_config_item):
         """
-        Determines pairs of input and output repos and also find correct source and debuginfo
-        counterpart of repos.
+        Searches for ubi repository triplet (binary rpm, srpm, debug) for
+        one ubi config item and tries to determine their population sources
+        (input repositories). Returns list UbiRepoSet objects that provides
+        input and output repositories that are used for population process.
         """
+        rpm_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs,
+                                             ubi_config_item.content_sets.rpm.output)
+        source_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs,
+                                                ubi_config_item.content_sets.srpm.output)
+        debug_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs,
+                                               ubi_config_item.content_sets.debuginfo.output)
 
-        rpms_cs = ubiconfig_item.content_sets.rpm
-        source_cs = ubiconfig_item.content_sets.srpm
-        debug_cs = ubiconfig_item.content_sets.debuginfo
-        _LOG.info(
-            "Getting input repos for input content sets:\n\t%s\n\t%s\n\t%s",
-            rpms_cs.input, source_cs.input, debug_cs.input)
+        ubi_repo_sets = []
 
-        in_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, rpms_cs.input)
-        in_source_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, source_cs.input)
-        in_debug_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, debug_cs.input)
+        for out_rpm_repo in rpm_repos_ft.result():
 
-        _LOG.info(
-            "Getting output repos for output content sets:\n\t%s\n\t%s\n\t%s",
-            rpms_cs.output, source_cs.output, debug_cs.output)
+            if not out_rpm_repo.ubi_population:
+                # it is sufficient to check only binary from repo triplet for disabling population
+                _LOG.debug(
+                    "Skipping population for output binary repo and "
+                    "related source and debug repos:\n\t%s",
+                    out_rpm_repo.repo_id)
+                continue
 
-        out_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, rpms_cs.output)
-        out_source_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, source_cs.output)
-        out_debug_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, debug_cs.output)
+            out_source_repo = self.get_repo_counterpart(out_rpm_repo, source_repos_ft.result())
+            out_debug_repo = self.get_repo_counterpart(out_rpm_repo, debug_repos_ft.result())
 
-        repo_pairs = []
-        for input_repo in in_repos_ft.result():
-            rpm = input_repo
-            source = self._get_repo_counterpart(input_repo, in_source_repos_ft.result())
-            debug_info = self._get_repo_counterpart(input_repo, in_debug_repos_ft.result())
+            in_rpm_repos = self._get_population_sources(out_rpm_repo,
+                                                        ubi_config_item.content_sets.rpm.input)
+            in_source_repos = self._get_population_sources(out_source_repo,
+                                                           ubi_config_item.content_sets.srpm.input)
+            in_debug_repos = self._get_population_sources(out_debug_repo,
+                                                          ubi_config_item.content_sets.debuginfo
+                                                          .input)
 
-            rhel_repo_set = RepoSet(rpm, source, debug_info)
+            out_repos = (out_rpm_repo, out_source_repo, out_debug_repo)
+            in_repos = (in_rpm_repos, in_source_repos, in_debug_repos)
 
-            rpm = self._get_repo_counterpart(input_repo, out_repos_ft.result())
-            source = self._get_repo_counterpart(input_repo, out_source_repos_ft.result())
-            debug_info = self._get_repo_counterpart(input_repo, out_debug_repos_ft.result())
+            ubi_repo_sets.append(UbiRepoSet(RepoSet(*in_repos), RepoSet(*out_repos)))
 
-            ubi_repo_set = RepoSet(rpm, source, debug_info)
+        return ubi_repo_sets
 
-            repo_pairs.append(UbiRepoSet(rhel_repo_set, ubi_repo_set))
+    def _get_population_sources(self, repo, input_cs):
+        src_repos = []
+        if repo.population_sources:
+            fts = [self._executor.submit(self.pulp.search_repo_by_id, r)
+                   for r in repo.population_sources]
 
-        return repo_pairs
+            for ft in as_completed(fts):
+                repo = ft.result()
+                if repo:
+                    src_repos.append(repo[0])
+        else:
+            in_repos_ft = self._executor.submit(self.pulp.search_repo_by_cs, input_cs)
+            repo = self.get_repo_counterpart(repo, in_repos_ft.result())
+            src_repos.append(repo)
 
-    def _get_repo_counterpart(self, input_repo, repos_to_match):
+        return src_repos
+
+    @staticmethod
+    def get_repo_counterpart(input_repo, repos_to_match):
+        """
+        Finds counterpart of input_repo in repos_to_match list by arch and platform_full_version.
+        """
         for repo in repos_to_match:
             if input_repo.arch == repo.arch and \
                     input_repo.platform_full_version == repo.platform_full_version:
@@ -181,13 +310,13 @@ class UbiPopulateRunner(object):
 
     def _match_modules(self):
         # Add matching modules
-
         fts = {}
         for module in self.ubiconfig.modules:
-            fts[self._executor.submit(self.pulp.search_modules,
-                                      self.repos.in_repos.rpm, module.name,
-                                      str(module.stream))] = \
-                (module.name + str(module.stream), module.profiles)
+            for in_repo_rpm in self.repos.in_repos.rpm:
+                fts[self._executor.submit(self.pulp.search_modules,
+                                          in_repo_rpm, module.name,
+                                          str(module.stream))] = \
+                    (module.name + str(module.stream), module.profiles)
 
         for ft in as_completed(fts):
             input_modules = ft.result()
@@ -226,29 +355,50 @@ class UbiPopulateRunner(object):
         fts = {}
         for _, modules in self.repos.modules.items():
             for md in modules:
-                fts[self._executor.submit(self.pulp.search_module_defaults,
-                                          self.repos.in_repos.rpm, md.name,
-                                          str(md.stream))] = md.name + str(md.stream)
+                for in_repo_rpm in self.repos.in_repos.rpm:
+                    fts[self._executor.submit(self.pulp.search_module_defaults,
+                                              in_repo_rpm, md.name,
+                                              str(md.stream))] = md.name + str(md.stream)
         for ft in as_completed(fts):
             module_defaults = ft.result()
             if module_defaults:
                 self.repos.module_defaults[fts[ft]].extend(module_defaults)
 
-    def _match_packages(self, repo, packages_dict):
+    def _get_pkgs_from_all_modules(self):
+        modules = []
+        for in_repo_rpm in self.repos.in_repos.rpm:
+            modules.extend(self.pulp.search_modules(in_repo_rpm))
+        pkgs = set()
+        regex = r'\d+:'
+        reg = re.compile(regex)
+        for module in modules:
+            for pkg in module.packages:
+                rpm_without_epoch = reg.sub('', pkg)
+                rpm_filename = rpm_without_epoch + '.rpm'
+                pkgs.add(rpm_filename)
+
+        return pkgs
+
+    def _match_packages(self, input_repos, packages_dict):
         """
         Add matching packages from whitelist
         Globbing package name is not supported
         """
+        modular_pkgs = self._get_pkgs_from_all_modules()
         fts = {}
         for package_pattern in self.ubiconfig.packages.whitelist:
             name = package_pattern.name
             arch = None if package_pattern.arch in ('*', None) else package_pattern.arch
-            fts[(self._executor.submit(self.pulp.search_rpms,
-                                       repo, name, arch))] = name
+            for repo in input_repos:
+                fts[(self._executor.submit(self.pulp.search_rpms,
+                                           repo, name, arch))] = name
 
         for ft in as_completed(fts):
             packages = ft.result()
             if packages:
+                for pkg in packages:
+                    if pkg.filename in modular_pkgs:
+                        pkg.is_modular = True
                 packages_dict[fts[ft]].extend(packages)
 
     def _match_binary_rpms(self):
@@ -324,16 +474,27 @@ class UbiPopulateRunner(object):
             if package.sourcerpm_filename is None:
                 _LOG.warning("Package %s doesn't reference its source rpm", package.name)
                 continue
+            in_repo = [r for r in self.repos.in_repos.rpm
+                       if r.repo_id == package.associate_source_repo_id][0]
+
+            associate_src_repo = UbiPopulate.get_repo_counterpart(in_repo,
+                                                                  self.repos.in_repos.source)
 
             self.repos.source_rpms[package.name].append(
-                Package(package.name, package.sourcerpm_filename)
+                Package(
+                    package.name,
+                    package.sourcerpm_filename,
+                    is_modular=package.is_modular,
+                    src_repo_id=associate_src_repo.repo_id)
             )
 
         blacklisted_srpms = self.get_blacklisted_packages(
-            list(chain.from_iterable(self.repos.source_rpms.values())))
+            list(chain.from_iterable(self.repos.source_rpms.values()))
+        )
 
         for pkg in blacklisted_srpms:
-            self.repos.source_rpms.pop(pkg.name, None)
+            if not pkg.is_modular:
+                self.repos.source_rpms.pop(pkg.name, None)
 
     def _determine_pulp_actions(self, units, current, diff_f):
         expected = list(chain.from_iterable(units.values()))
@@ -354,6 +515,25 @@ class UbiPopulateRunner(object):
     def _get_pulp_actions_pkgs(self, pkgs, current):
         return self._determine_pulp_actions(pkgs, current, self._diff_packages_by_filename)
 
+    def _get_pulp_actions_src_pkgs(self, pkgs, current):
+        """
+        Get required pulp actions to make sure existing and desired source packages are in
+        match.
+        """
+        src_pkgs = {}
+        uniq_srpms = {}
+        # filter out packages that share same source rpm
+        for _, pkgs in pkgs.items():
+            for pkg in pkgs:
+                fn = pkg.filename or pkg.sourcerpm_filename
+                uniq_srpms[fn] = pkg
+
+        # remap uniq srpms to format accepted by _determine_pulp_actions
+        for _, srpm in uniq_srpms.items():
+            src_pkgs[(srpm.name, srpm.version, srpm.release)] = [srpm]
+
+        return self._determine_pulp_actions(src_pkgs, current, self._diff_packages_by_filename)
+
     def _get_pulp_actions(self, current_modules_ft, current_module_defaults_ft, current_rpms_ft,
                           current_srpms_ft, current_debug_rpms_ft):
         """
@@ -372,8 +552,8 @@ class UbiPopulateRunner(object):
 
         rpms_assoc, rpms_unassoc = self._get_pulp_actions_pkgs(self.repos.packages,
                                                                current_rpms_ft.result())
-        srpms_assoc, srpms_unassoc = self._get_pulp_actions_pkgs(self.repos.source_rpms,
-                                                                 current_srpms_ft.result())
+        srpms_assoc, srpms_unassoc = self._get_pulp_actions_src_pkgs(self.repos.source_rpms,
+                                                                     current_srpms_ft.result())
 
         debug_assoc = None
         debug_unassoc = None
@@ -424,6 +604,18 @@ class UbiPopulateRunner(object):
 
         return diff
 
+    def _filter_pkgs_from_modules(self):
+        regex = r'\d+:'
+        reg = re.compile(regex)
+
+        for name_stream, packages in self.repos.pkgs_from_modules.items():
+            # name_stream, packages == str, list of _pulp_client.Package
+
+            wanted_packages = list(chain.from_iterable([m.packages for m in self.repos.modules[name_stream]]))
+            wanted_packages = [reg.sub('', rpm_nevra) + '.rpm' for rpm_nevra in wanted_packages]
+            packages[:] = [pkg for pkg in packages if pkg.filename in wanted_packages]
+
+
     def run_ubi_population(self):
         current_modules_ft, current_module_defaults_ft, current_rpms_ft, \
             current_srpms_ft, current_debug_rpms_ft = self._get_current_content()
@@ -434,6 +626,7 @@ class UbiPopulateRunner(object):
             self._match_debug_rpms()
         self._exclude_blacklisted_packages()
         self._finalize_modules_output_set()
+        self._filter_pkgs_from_modules()
         self._finalize_rpms_output_set()
         self._finalize_debug_output_set()
         self._match_module_defaults()
@@ -473,7 +666,7 @@ class UbiPopulateRunner(object):
         fts = []
         for action in action_list:
             if action.units:
-                fts.append(self._executor.submit(*action.get_action(self.pulp)))
+                fts.extend([self._executor.submit(*a) for a in action.get_actions(self.pulp)])
 
         return fts
 
@@ -522,12 +715,13 @@ class UbiPopulateRunner(object):
         for item in associations:
             if item.units:
                 for unit in item.units:
-                    _LOG.info("Would associate %s from %s to %s", unit, item.src_repo.repo_id,
+                    _LOG.info("Would associate %s from %s to %s", unit,
+                              unit.associate_source_repo_id,
                               item.dst_repo.repo_id)
 
             else:
                 _LOG.info("No association expected for %s from %s to %s", item.TYPE,
-                          item.src_repo.repo_id,
+                          [r.repo_id for r in item.src_repos],
                           item.dst_repo.repo_id)
 
         for item in unassociations:
@@ -605,7 +799,14 @@ class UbiPopulateRunner(object):
         """
         Keeps n latest modules in modules sorted list
         """
-        modules[:] = modules[-n:]
+        modules_to_keep = []
+        versions_to_keep = sorted(set([m.version for m in modules]))[-n:]
+
+        for module in modules:
+            if module.version in versions_to_keep:
+                modules_to_keep.append(module)
+
+        modules[:] = modules_to_keep
 
     def sort_packages(self, packages):
         """
@@ -660,56 +861,78 @@ class UbiPopulateRunner(object):
                     continue
                 rpm_filename = rpm_without_epoch + '.rpm'
 
-                # Check existence of rpm in binary rpm repo
-                rpms = self.pulp.search_rpms(self.repos.in_repos.rpm, filename=rpm_filename)
+                # Check existence of rpm in binary rpm repos
+                rpms = []
 
+                for in_repo_rpm in self.repos.in_repos.rpm:
+                    res = self.pulp.search_rpms(in_repo_rpm, filename=rpm_filename)
+                    if res:
+                        rpms.extend(res)
                 if rpms:
+                    rpms[0].is_modular = True
                     ret_rpms.append(rpms[0])
                 else:
-                    # Check existence of rpm in debug repo
-                    debug_rpms = self.pulp.search_rpms(
-                        self.repos.in_repos.debug,
-                        filename=rpm_filename,
-                    )
+                    # Check existence of rpm in debug repos
+                    debug_rpms = []
+                    for in_repo_debug in self.repos.in_repos.debug:
+                        res = self.pulp.search_rpms(in_repo_debug, filename=rpm_filename)
+                        if res:
+                            debug_rpms.extend(res)
 
                     if debug_rpms:
+                        debug_rpms[0].is_modular = True
                         ret_debug_rpms.append(debug_rpms[0])
                     else:
                         _LOG.warning("RPM %s is unavailable in input repos %s %s, skipping",
-                                     rpm_filename, self.repos.in_repos.rpm.repo_id,
-                                     self.repos.in_repos.debug.repo_id
+                                     rpm_filename,
+                                     [r.repo_id for r in self.repos.in_repos.rpm],
+                                     [r.repo_id for r in self.repos.in_repos.debug]
                                      )
         return ret_rpms, ret_debug_rpms
 
     def keep_n_latest_packages(self, packages, n=1):
         """
-        Keep n latest non-modular packages,
-        modular packages are kept only if they are referenced by some of remaining modules
-        Parameter packages: sorted list of Packages objects, oldest goes first
+        Keep n latest non-modular packages.
+        Modular packages are kept only if they are referenced by some of
+        remaining modules.
+
+        Arguments:
+            packages (List[Package]): Sorted, oldest goes first
+
+        Keyword arguments:
+            n (int): Number of non-modular package versions to keep
+
+        Returns:
+            None. The packages list is changed in-place
         """
+
         packages_to_keep = []
-        non_modular_pkgs = []
+        filenames_to_keep = set()  # from modular packages
+
+        # Use a queue of n elements per arch
+        pkgs_per_arch = defaultdict(lambda: deque(maxlen=n))
+
+        # Set of package filenames from modules. Modular packages in
+        # packages list are kept if referenced here.
+        modular_packages_filenames = set()
+        for module_name_stream, packages_in_module in self.repos.pkgs_from_modules.items():
+            if module_name_stream in self.repos.modules:
+                modular_packages_filenames |= set([pkg.filename for pkg in packages_in_module])
+
         for package in packages:
-            if package.is_modular:
-                for module_name_stream, packages_ref_by_module in \
-                        self.repos.pkgs_from_modules.items():
-                    pkgs_filenames_from_modules = [pkg.filename for pkg in packages_ref_by_module]
-                    if package.filename in pkgs_filenames_from_modules \
-                            and module_name_stream in self.repos.modules:
-                        # this skips modular pkgs that are not referenced by module
-                        packages_to_keep.append(package)
-            else:
-                non_modular_pkgs.append(package)
+            if not package.is_modular:
+                # filter non-modular pkgs per arches, there can be rpms
+                # with different arches for package in one repository
+                _, _, _, _, arch = split_filename(package.filename)
+                pkgs_per_arch[arch].append(package)
+            elif (package.filename in modular_packages_filenames
+                  and package.filename not in filenames_to_keep):
+                # this skips modular pkgs that are not referenced by module
+                packages_to_keep.append(package)
+                filenames_to_keep.add(package.filename)
 
-        # filter non-modular pkgs per arches, there can be rpms with different arches
-        # for package in one repository
-        pkgs_per_arch = defaultdict(list)
-        for pkg in non_modular_pkgs:
-            _, _, _, _, arch = split_filename(pkg.filename)
-            pkgs_per_arch[arch].append(pkg)
-
-        latest_pkgs_per_arch = []
-        for pkgs in pkgs_per_arch.values():
-            latest_pkgs_per_arch += pkgs[-n:]
+        # Packages already included from modules are skipped
+        latest_pkgs_per_arch = [pkg for pkg in chain.from_iterable(pkgs_per_arch.values())
+                                if pkg.filename not in filenames_to_keep]
 
         packages[:] = latest_pkgs_per_arch + packages_to_keep
